"""HTTP User Agents Example"""

import argparse
import os
import sys
from collections import Counter
from pprint import pprint

# Local imports
from zat import zeek_log_reader

if __name__ == "__main__":
    # Example to run the zeek log reader on a given file

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument("zeek_log", type=str, help="Specify a zeek log to run ZeekLogReader test on")
    parser.add_argument("-t", action="store_true", default=False, help="Sets the program to tail a live Zeek log")
    parser.add_argument("-s", action="store_true", default=False, help="Only print the summary of the findings.")
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print("Unrecognized args: %s" % commands)
        sys.exit(1)

    # Sanity check that this is a http log
    if "http" not in args.zeek_log:
        print("This example only works with Zeek http.log files..")
        sys.exit(1)

    # File may have a tilde in it
    if args.zeek_log:
        args.zeek_log = os.path.expanduser(args.zeek_log)

        # Run the zeek reader on a given log file counting up user agents
        http_agents = Counter()
        reader = zeek_log_reader.ZeekLogReader(args.zeek_log, tail=args.t)
        for count, row in enumerate(reader.readrows()):
            # Track count
            http_agents[row["user_agent"]] += 1

            # Every 10000 rows report agent counts (least common)
            if not args.s:
                if count % 10000 == 0:
                    print("\n<<<Least Common User Agents>>>")
                    pprint(http_agents.most_common()[:-50:-1])

        # Also report at the end (if there is one)
        print("\nMost Common User Agents:")
        pprint(http_agents.most_common()[:20])

        print("\nLeast Common User Agents:")
        pprint(http_agents.most_common()[:-50:-1])



"""LogToDataFrame: Converts a Zeek log to a Pandas DataFrame"""

# Third Party
import pandas as pd

# Local
from zat.utils.field_info import get_field_info


class LogToDataFrame:
    """LogToDataFrame: Converts a Zeek log to a Pandas DataFrame
    Notes:
        This class has recently been overhauled from a simple loader to a more
        complex class that should in theory:
          - Select better types for each column
          - Should be faster
          - Produce smaller memory footprint dataframes
        If you have any issues/problems with this class please submit a GitHub issue.
    More Info: https://supercowpowers.github.io/zat/large_dataframes.html
    """

    def __init__(self):
        """Initialize the LogToDataFrame class"""

        # First Level Type Mapping
        #    This map defines the types used when first reading in the Zeek log into a 'chunk' dataframes.
        #    Types (like time and interval) will be defined as one type at first but then
        #    will undergo further processing to produce correct types with correct values.
        # See: https://stackoverflow.com/questions/29245848/what-are-all-the-dtypes-that-pandas-recognizes
        #      for more info on supported types.
        self.type_map = {
            "bool": "category",  # Can't hold NaN values in 'bool', so we're going to use category
            "count": "UInt64",
            "int": "Int32",
            "double": "float",
            "time": "float",  # Secondary Processing into datetime
            "interval": "float",  # Secondary processing into timedelta
            "port": "UInt16",
        }

    def _create_initial_df(self, log_filename, all_fields, usecols, dtypes):
        """Internal Method: Create the initial dataframes by using Pandas read CSV (primary types correct)"""
        return pd.read_csv(
            log_filename, sep="\t", names=all_fields, usecols=usecols, dtype=dtypes, comment="#", na_values="-"
        )

    def create_dataframe(self, log_filename, ts_index=True, aggressive_category=True, usecols=None):
        """Create a Pandas dataframe from a Bro/Zeek log file
        Args:
           log_fllename (string): The full path to the Zeek log
           ts_index (bool): Set the index to the 'ts' field (default = True)
           aggressive_category (bool): convert unknown columns to category (default = True)
           usecol (list): A subset of columns to read in (minimizes memory usage) (default = None)
        """

        # Grab the field information
        field_names, field_types = get_field_info(log_filename=log_filename)
        all_fields = field_names

        # If usecols is set then we'll subset the fields and types
        if usecols:
            # Usecols needs to include ts
            if "ts" not in usecols:
                usecols.append("ts")
            field_types = [t for t, field in zip(field_types, field_names) if field in usecols]
            field_names = [field for field in field_names if field in usecols]

        # Get the appropriate types for the Pandas Dataframe
        type_map = self.pd_column_types(field_names, field_types, aggressive_category)

        # Now actually read in the initial dataframe
        self._df = self._create_initial_df(
            log_filename=log_filename, all_fields=all_fields, usecols=usecols, dtypes=type_map
        )

        # Now we convert 'time' and 'interval' fields to datetime and timedelta respectively
        for name, zeek_type in zip(field_names, field_types):
            if zeek_type == "time":
                self._df[name] = pd.to_datetime(self._df[name], unit="s")
            if zeek_type == "interval":
                self._df[name] = pd.to_timedelta(self._df[name], unit="s")

        # Set the index
        if ts_index and not self._df.empty:
            try:
                self._df.set_index("ts", inplace=True)
            except KeyError:
                print("Could not find ts/timestamp for index...")
        return self._df

    def pd_column_types(self, column_names, column_types, aggressive_category=True, verbose=False):
        """Given a set of names and types, construct a dictionary to be used
        as the Pandas read_csv dtypes argument"""

        # Aggressive Category means that types not in the current type_map are
        # mapped to a 'category' if aggressive_category is False then they
        # are mapped to an 'object' type
        unknown_type = "category" if aggressive_category else "object"

        pandas_types = {}
        for name, zeek_type in zip(column_names, column_types):

            # Grab the type
            item_type = self.type_map.get(zeek_type)

            # Sanity Check
            if not item_type:
                # UID/FUID/GUID always gets mapped to object
                if "uid" in name:
                    item_type = "object"
                else:
                    if verbose:
                        print("Could not find type for {:s} using {:s}...".format(zeek_type, unknown_type))
                    item_type = unknown_type

            # Set the pandas type
            pandas_types[name] = item_type

        # Return the dictionary of name: type
        return pandas_types


# Simple test of the functionality
def test():
    """Test for LogToDataFrame Class"""
    import os

    pd.set_option("display.width", 1000)
    from zat.utils import file_utils

    # Grab a test file
    data_path = file_utils.relative_dir(__file__, "../data")
    log_path = os.path.join(data_path, "conn.log")

    # Convert it to a Pandas DataFrame
    log_to_df = LogToDataFrame()
    my_df = log_to_df.create_dataframe(log_path)

    # Print out the head
    print(my_df.head())

    # Print out the datatypes
    print(my_df.dtypes)

    # Test a bunch
    tests = [
        "app_stats.log",
        "dns.log",
        "http.log",
        "notice.log",
        "tor_ssl.log",
        "conn.log",
        "dhcp_002.log",
        "files.log",
        "smtp.log",
        "weird.log",
        "ftp.log",
        "ssl.log",
        "x509.log",
    ]
    for log_path in [os.path.join(data_path, log) for log in tests]:
        print("Testing: {:s}...".format(log_path))
        my_df = log_to_df.create_dataframe(log_path)
        print(my_df.head())
        print(my_df.dtypes)

    # Test out usecols arg
    conn_path = os.path.join(data_path, "conn.log")
    my_df = log_to_df.create_dataframe(
        conn_path, usecols=["id.orig_h", "id.orig_p", "id.resp_h", "id.resp_p", "proto", "orig_bytes", "resp_bytes"]
    )

    # Test an empty log (a log with header/close but no data rows)
    log_path = os.path.join(data_path, "http_empty.log")
    my_df = log_to_df.create_dataframe(log_path)

    # Print out the head
    print(my_df.head())

    # Print out the datatypes
    print(my_df.dtypes)

    print("LogToDataFrame Test successful!")


if __name__ == "__main__":
    # Run the test for easy testing/debugging
    test()



"""Zeek log to Parquet Dataframe Example"""

import argparse
import os
import sys

# Note: We're going to import pyarrow but it currently has an open issue around supporting time deltas
#    - https://issues.apache.org/jira/browse/ARROW-6780 so we have to convert timedelta fields to str
# Also see:
#   - https://stackoverflow.com/questions/53893554/transfer-and-write-parquet-with-python-and-pandas-got-timestamp-error
#
from datetime import timedelta

import pandas as pd

try:
    import pyarrow  # noqa: F401
except ImportError:
    print("Please > pip install pyarrow")
    sys.exit(1)

# Local imports
from zat.log_to_dataframe import LogToDataFrame


# Helper method for temporarily converting timedelta to string
def convert_timedelta_to_str(df):
    delta_columns = df.select_dtypes(include=["timedelta"])
    for column in delta_columns:
        df[column] = df[column].apply(tdelta_value_to_str)
    return df


def tdelta_value_to_str(value):
    if pd.isnull(value):
        return "-"  # Standard for Zeek null value
    else:
        return str(timedelta(seconds=value.total_seconds()))


if __name__ == "__main__":
    # Example to write Parquet file from a zeek log

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument("zeek_log", type=str, help="Specify the zeek log input file")
    parser.add_argument("parquet_file", type=str, help="Specify the parquet file to write out")
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print("Unrecognized args: %s" % commands)
        sys.exit(1)

    # File may have a tilde in it
    if args.zeek_log and args.parquet_file:
        args.zeek_log = os.path.expanduser(args.zeek_log)
        args.parquet_file = os.path.expanduser(args.parquet_file)

        # Convert to dataframe and write out the parquet file
        log_to_df = LogToDataFrame()
        zeek_df = log_to_df.create_dataframe(args.zeek_log)
        print("Dataframe Created: {:d} rows...".format(len(zeek_df)))

        # Check for any timedelta fields (see note above)
        df = convert_timedelta_to_str(zeek_df)

        zeek_df.to_parquet(args.parquet_file, compression="snappy", use_deprecated_int96_timestamps=True)
        print("Complete: {:s} --> {:s}".format(args.zeek_log, args.parquet_file))



"""Anomaly Detection Example"""

import argparse
import math
import os
import sys
from collections import Counter

# Third Party Imports
import pandas as pd
from sklearn.cluster import KMeans
from sklearn.ensemble import IsolationForest

# Local imports
from zat import dataframe_to_matrix, log_to_dataframe


def entropy(string):
    """Compute entropy on the string"""
    p, lns = Counter(string), float(len(string))
    return -sum(count / lns * math.log(count / lns, 2) for count in p.values())


if __name__ == "__main__":
    # Example to show the dataframe cache functionality on streaming data
    pd.set_option("display.width", 1000)

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument("zeek_log", type=str, help="Specify a zeek log to run ZeekLogReader test on")
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print("Unrecognized args: %s" % commands)
        sys.exit(1)

    # File may have a tilde in it
    if args.zeek_log:
        args.zeek_log = os.path.expanduser(args.zeek_log)

        # Sanity check either http or dns log
        if "http" in args.zeek_log:
            log_type = "http"
            features = ["id.resp_p", "method", "resp_mime_types", "request_body_len"]
        elif "dns" in args.zeek_log:
            log_type = "dns"
            features = ["Z", "proto", "qtype_name", "query_length", "answer_length", "entropy"]
        else:
            print("This example only works with Zeek with http.log or dns.log files..")
            sys.exit(1)

        # Create a Pandas dataframe from a Zeek log
        try:
            log_to_df = log_to_dataframe.LogToDataFrame()
            zeek_df = log_to_df.create_dataframe(args.zeek_log)
            print(zeek_df.head())
        except IOError:
            print("Could not open or parse the specified logfile: %s" % args.zeek_log)
            sys.exit(1)
        print("Read in {:d} Rows...".format(len(zeek_df)))

        # Using Pandas we can easily and efficiently compute additional data metrics
        # Here we use the vectorized operations of Pandas/Numpy to compute query length
        # We'll also compute entropy of the query
        if log_type == "dns":
            zeek_df["query_length"] = zeek_df["query"].str.len()
            zeek_df["answer_length"] = zeek_df["answers"].str.len()
            zeek_df["entropy"] = zeek_df["query"].map(lambda x: entropy(x))

        # Use the zat DataframeToMatrix class
        to_matrix = dataframe_to_matrix.DataFrameToMatrix()
        zeek_matrix = to_matrix.fit_transform(zeek_df[features])
        print(zeek_matrix.shape)

        # Train/fit and Predict anomalous instances using the Isolation Forest model
        odd_clf = IsolationForest(contamination=0.2)  # Marking 20% as odd
        odd_clf.fit(zeek_matrix)

        # Now we create a new dataframe using the prediction from our classifier
        predictions = odd_clf.predict(zeek_matrix)
        odd_df = zeek_df[features][predictions == -1]
        display_df = zeek_df[predictions == -1].copy()

        # Now we're going to explore our odd observations with help from KMeans
        odd_matrix = to_matrix.fit_transform(odd_df)
        num_clusters = min(len(odd_df), 4)  # 4 clusters unless we have less than 4 observations
        display_df["cluster"] = KMeans(n_clusters=num_clusters).fit_predict(odd_matrix)
        print(odd_matrix.shape)

        # Now group the dataframe by cluster
        if log_type == "dns":
            features += ["query"]
        else:
            features += ["host"]
        cluster_groups = display_df[features + ["cluster"]].groupby("cluster")

        # Now print out the details for each cluster
        print("<<< Outliers Detected! >>>")
        for key, group in cluster_groups:
            print("\nCluster {:d}: {:d} observations".format(key, len(group)))
            print(group.head())



"""DataFrameToMatrix: Class that converts a DataFrame to a Numpy Matrix (ndarray)"""

# Third Party
import numpy as np
import pandas as pd


class DataFrameToMatrix:
    """DataFrameToMatrix: Class that converts a DataFrame to a Numpy Matrix (ndarray)
    Notes:
        fit_transform: Does a fit and a transform and returns the transformed matrix
        transform: Based on previous fit parameters returns the transformed matrix
    More Info: https://supercowpowers.github.io/zat/dataframe_to_matrix.html

    # Nullable integer arrays are currently not handled by Numpy
    # Cast Nullable integer arrays to float32
    null_int_types = [pd.UInt16Dtype, pd.UInt32Dtype, pd.UInt64Dtype, pd.Int64Dtype]
    for col in _internal_df:
        if type(_internal_df[col].dtype) in null_int_types:
            _internal_df[col] = _internal_df[col].astype(np.float32)

    """

    def __init__(self):
        """Initialize the DataFrameToMatrix class"""
        self.column_names = None
        self.norm_map = {}
        self.dtype_map = {}
        self.nan_replace = -999

    def fit_transform(self, input_df, normalize=True, nan_replace=-999, copy=True):
        """Convert the dataframe to a matrix (numpy ndarray)
        Args:
            input_df (dataframe): The dataframe to convert
            normalize (bool): Boolean flag to normalize numeric columns (default=True)
        """
        self.nan_replace = nan_replace

        # Copy the dataframe (if wanted)
        _internal_df = input_df.copy() if copy else input_df

        # Convert object columns to categorical
        self.object_to_categorical(_internal_df)

        # Convert categorical NaNs
        self.fit_category_nans(_internal_df)

        # Lock categories to specific values (important for train/predict consistency)
        self.lock_categorical(_internal_df)

        # Sanity Check
        self.sanity_check_categorical(_internal_df)

        # Normalize numeric columns (mean normalize, sometimes called 'standardizing')
        if normalize:
            self.normalize_numeric(_internal_df)

        # Remove any numerical NaNs (categorical NaNs were removed above)
        for column in _internal_df.select_dtypes(include="number").columns:
            _internal_df[column].fillna(self.nan_replace, inplace=True)

        # Drop any columns that aren't numeric or categorical
        for column in list(_internal_df.select_dtypes(exclude=["number", "category"]).columns):
            print("Dropping {:s} column...".format(column))
        _internal_df = _internal_df.select_dtypes(include=["number", "category"])

        # Capture all the column/dtype information from the dataframe
        self.column_names = _internal_df.columns.to_list()
        for column in _internal_df.columns:
            self.dtype_map[column] = _internal_df[column].dtype

        # Now with every thing setup, call the dummy_encoder, convert to ndarray and return
        return pd.get_dummies(_internal_df).to_numpy(dtype=np.float32)

    def transform(self, input_df, copy=True):
        """Convert the dataframe to a matrix (numpy ndarray)
        Args:
            input_df (dataframe): The dataframe to convert
        """

        # Make sure we have the same columns in the right order
        # Copy the dataframe (if wanted)
        _internal_df = input_df[self.column_names].copy() if copy else input_df[self.column_names]

        # Convert all columns to the proper dtypes
        for column, dtype in self.dtype_map.items():
            _internal_df[column] = _internal_df[column].astype(dtype)

        # Convert any categorical NaNs to a 'NaN' category
        self.transform_category_nans(_internal_df)

        # Normalize any numeric columns
        for column, (smin, smax) in self.norm_map.items():
            print("Normalizing column {:s}...".format(column))
            _internal_df[column] = (_internal_df[column] - smin) / (smax - smin)

        # Remove any numerical NaNs (categorical NaNs were removed above)
        for column in _internal_df.select_dtypes(include="number").columns:
            _internal_df[column].fillna(self.nan_replace, inplace=True)

        # Now with every thing setup, call the dummy_encoder, convert to ndarray and return
        return pd.get_dummies(_internal_df).to_numpy(dtype=np.float32)

    @staticmethod
    def fit_category_nans(df):
        """ONLY FIT: Convert np.NaNs to a category 'NaN'"""
        for column in df.select_dtypes(include=["category"]).columns:
            if df[column].isnull().any():
                df[column] = df[column].cat.add_categories("NaN")
                df[column].fillna("NaN", inplace=True)

    @staticmethod
    def transform_category_nans(df):
        """ONLY TRANSFORM: Convert np.NaNs to a category 'NaN'"""
        for column in df.select_dtypes(include=["category"]).columns:
            if "NaN" in df[column].cat.categories:
                df[column].fillna("NaN", inplace=True)

    @staticmethod
    def object_to_categorical(df):
        """Run a heuristic on the object columns to determine whether it contains categorical values
           if the heuristic decides it's categorical then the type of the column is changed
        Args:
            df (dataframe): The dataframe to check for categorical data
        Returns:
            None but the dataframe columns are modified
        """

        # Loop through each column that might be converable to categorical
        for column in df.select_dtypes(include="object").columns:

            # If we don't have too many unique values convert the column
            if df[column].nunique() < 100:
                print("Changing column {:s} to category...".format(column))
                df[column] = pd.Categorical(df[column])

    @staticmethod
    def lock_categorical(df):
        """Lock the categorical column types to a specific ordered list of categories
        Args:
            df (dataframe): The dataframe to lock categorical columns
        Returns:
            None but note that the dataframe is modified to 'lock' the categorical columns
        """
        for column in df.select_dtypes(include="category").columns:
            cats = [c for c in df[column].unique().tolist() if pd.notna(c)]
            df[column] = pd.Categorical(df[column], categories=sorted(cats))

    @staticmethod
    def sanity_check_categorical(df):
        """Sanity check for 'dimensionality explosion' on categorical types
        Args:
            df (dataframe): The dataframe to check the categorical columns
        Returns:
            None
        """
        for column in df.select_dtypes(include="category").columns:
            # Give warning on category types will LOTs of values
            num_unique = df[column].nunique()
            if num_unique > 20:
                print("WARNING: {:s} will expand into {:d} dimensions...".format(column, num_unique))

    def normalize_numeric(self, df):
        """Normalize (mean normalize) the numeric columns in the dataframe
        Args:
            df (dataframe): The dataframe to normalize
        Returns:
            None but note that the numeric columns of the dataframe are modified
        """
        for column in df.select_dtypes(include="number").columns:
            print("Normalizing column {:s}...".format(column))
            df[column] = self._normalize_series(df[column])

    def _normalize_series(self, series):
        smin = series.min()
        smax = series.max()

        # Check for div by 0
        if smax - smin == 0:
            print("Cannot normalize series (div by 0) so not normalizing...")
            return series

        # Capture the normalization info and return the normalize series
        self.norm_map[series.name] = (smin, smax)
        return (series - smin) / (smax - smin)


# Simple test of the functionality
def test():
    """Test for DataFrameToMatrix Class"""
    import os
    import pickle
    from tempfile import NamedTemporaryFile

    import numpy.testing as np_test_utils

    pd.set_option("display.width", 1000)

    test_df = pd.DataFrame(
        {
            "A": pd.Categorical(["a", "b", "c", "a"], ordered=True),
            "B": pd.Categorical(["a", "b", "c", "a"], ordered=False),
            "C": pd.Categorical(["a", "b", "z", "a"], categories=["a", "b", "z", "d"]),
            "D": [1, 2, 3, 4],
            "E": ["w", "x", "y", "z"],
            "F": [1.1, 2.2, 3.3, 4.4],
            "G": pd.to_datetime([0, 1, 2, 3]),
            "H": [True, False, False, True],
        }
    )
    test_df2 = pd.DataFrame(
        {
            "A": pd.Categorical(["a", "b", "b", "a"], ordered=True),
            "B": pd.Categorical(["a", "b", "d", "a"], ordered=False),
            "C": pd.Categorical(["a", "b", "z", "y"], categories=["a", "b", "z", "d"]),
            "D": [1, 2, 3, 7],
            "E": ["w", "x", "z", "foo"],
            "F": [1.1, 2.2, 3.3, 4.4],
            "H": [True, False, False, False],
        }
    )

    # Copy the test_df for testing later
    copy_test_df = test_df.copy()

    # Test the transformation from dataframe to numpy ndarray and back again
    to_matrix = DataFrameToMatrix()
    print("FIT-TRANSFORM")
    matrix = to_matrix.fit_transform(test_df)
    print("TRANSFORM")
    matrix_test = to_matrix.transform(test_df)

    # These two matrices should be the same
    np_test_utils.assert_equal(matrix, matrix_test)

    # Assert that the dataframe we passed in didn't change
    copy_test_df.equals(test_df)

    # Test that the conversion gives us the same columns on a df with different category values
    # This also tests NaN in a category column
    print("TRANSFORM2")
    matrix2 = to_matrix.transform(test_df2)
    assert matrix.shape == matrix2.shape

    # First two ROWS should be the same
    np_test_utils.assert_equal(matrix[0], matrix2[0])
    np_test_utils.assert_equal(matrix[1], matrix2[1])

    # Test normalize
    to_matrix_norm = DataFrameToMatrix()
    print("FIT-TRANSFORM")
    norm_matrix = to_matrix_norm.fit_transform(test_df)
    print(norm_matrix)
    assert norm_matrix[:, 0].min() == 0
    assert norm_matrix[:, 0].max() == 1

    # Make sure normalize 'does the right thing' when doing transform
    print("TRANSFORM")
    norm_matrix2 = to_matrix_norm.transform(test_df2)
    assert norm_matrix2[:, 0].min() == 0
    assert norm_matrix2[:, 0].max() == 2  # Normalization is based on FIT range

    # Test div by zero in normalize
    test_df3 = test_df2.copy()
    test_df3["D"] = [1, 1, 1, 1]
    print("FIT-TRANSFORM")
    norm_matrix3 = to_matrix_norm.fit_transform(test_df3)
    assert norm_matrix3[:, 0].min() == 1
    assert norm_matrix3[:, 0].max() == 1

    # Test serialization
    temp = NamedTemporaryFile(delete=False)
    pickle.dump(to_matrix, temp)
    temp.close()

    # Deserialize and test
    to_matrix_from_disk = pickle.load(open(temp.name, "rb"))
    print("TRANSFORM")
    matrix3 = to_matrix_from_disk.transform(test_df)
    print("TRANSFORM")
    matrix4 = to_matrix_from_disk.transform(test_df2)
    np_test_utils.assert_equal(matrix, matrix3)
    np_test_utils.assert_equal(matrix2, matrix4)

    # Remove the temporary file
    os.unlink(temp.name)

    # Try 'nullable' integer arrays
    null_df = test_df2.copy()
    null_df["I"] = pd.Series([10, 11, 12, np.nan], dtype="UInt64")
    print("FIT-TRANSFORM")
    matrix = to_matrix.fit_transform(null_df)
    print("TRANSFORM")
    matrix_test = to_matrix.transform(null_df)

    # These two matrices should be the same
    np_test_utils.assert_equal(matrix, matrix_test)

    # Now actually try the matrix with a scikit-learn algo
    from sklearn.cluster import KMeans

    to_matrix = DataFrameToMatrix()
    my_matrix = to_matrix.fit_transform(test_df)
    kmeans = KMeans(n_clusters=2).fit_predict(my_matrix)

    # Now we can put our ML results back onto our dataframe!
    test_df["cluster"] = kmeans
    cluster_groups = test_df.groupby("cluster")

    # Now print out the details for each cluster
    for _key, group in cluster_groups:
        print("Rows in Cluster: {:d}".format(len(group)))
        print(group.head(), "\n")
    del test_df["cluster"]

    # Now we're going to intentionally introduce NaNs in the categorical output just to see what happens
    to_matrix = DataFrameToMatrix()
    _ = to_matrix.fit_transform(test_df)
    my_matrix2 = to_matrix.transform(test_df2)
    kmeans = KMeans(n_clusters=2).fit_predict(my_matrix2)

    # Now we can put our ML results back onto our dataframe!
    test_df2["cluster"] = kmeans
    cluster_groups = test_df2.groupby("cluster")

    # Now print out the details for each cluster
    for _key, group in cluster_groups:
        print("Rows in Cluster: {:d}".format(len(group)))
        print(group.head(), "\n")

    # Test a corner case (submitted via GitHub)
    cat_df = pd.DataFrame({"a": ["a", "b", "c"], "b": ["v", "c", "a"]})
    cat_df = cat_df.astype("category")
    copy_cat_df = cat_df.copy()
    print("FIT-TRANSFORM")
    to_matrix = DataFrameToMatrix()
    matrix = to_matrix.fit_transform(cat_df)
    print(matrix)
    print("TRANSFORM")
    matrix_test = to_matrix.transform(cat_df)
    print(matrix_test)

    # These two matrices should be the same
    np_test_utils.assert_equal(matrix, matrix_test)

    # Assert that the dataframe we passed in didn't change
    copy_cat_df.equals(cat_df)


if __name__ == "__main__":
    test()



"""Signal Catching utility"""

import signal
import sys
import time
from contextlib import contextmanager


@contextmanager
def signal_catcher(callback, **kwargs):
    """Catch signals and invoke the callback method"""

    def _catch_exit_signal(sig_num, _frame):
        print("Received signal {:d} invoking callback...".format(sig_num))
        callback(**kwargs)

    signal.signal(signal.SIGINT, _catch_exit_signal)
    signal.signal(signal.SIGQUIT, _catch_exit_signal)
    signal.signal(signal.SIGTERM, _catch_exit_signal)
    yield


def my_exit():
    """My exit callback (just for testing)"""
    print("My Exit got called...Do some cleanup or whatever...")
    sys.exit()


def test():
    """Test the SignalCatcher"""
    with signal_catcher(my_exit):
        time.sleep(1)


if __name__ == "__main__":
    test()



"""Risky DNS zat Example"""

import argparse
import os
import pickle
import sys
from pprint import pprint

# Third Party Imports
try:
    import tldextract
except ImportError:
    print("\nThis example needs tldextract. Please do a $pip install tldextract and rerun this example")
    sys.exit(1)

# Local imports
from zat import zeek_log_reader
from zat.utils import signal_utils, vt_query


def save_vtq():
    """Exit on Signal"""
    global vtq  # noqa: F824

    print("Saving VirusTotal Query Cache...")
    pickle.dump(vtq, open("vtq.pkl", "wb"), protocol=pickle.HIGHEST_PROTOCOL)
    sys.exit()


if __name__ == "__main__":
    # Risky DNS/VT Query application

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument("zeek_log", type=str, help="Specify a zeek log to run ZeekLogReader test on")
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print("Unrecognized args: %s" % commands)
        sys.exit(1)

    # Sanity check that this is a dns log
    if "dns" not in args.zeek_log:
        print("This example only works with Zeek dns.log files..")
        sys.exit(1)

    # File may have a tilde in it
    if args.zeek_log:
        args.zeek_log = os.path.expanduser(args.zeek_log)

        # See if we have a serialized VirusTotal Query Class.
        # If we do not have one we'll create a new one
        try:
            vtq = pickle.load(open("vtq.pkl", "rb"))
            print("Opening VirusTotal Query Cache (cache_size={:d})...".format(vtq.size))
        except IOError:
            vtq = vt_query.VTQuery(max_cache_time=60 * 24 * 7)  # One week cache

        # See our 'Risky Domains' Notebook for the analysis and
        # statistical methods used to compute this risky set of TLDs
        risky_tlds = set(
            [
                "info",
                "tk",
                "xyz",
                "online",
                "club",
                "ru",
                "website",
                "in",
                "ws",
                "top",
                "site",
                "work",
                "biz",
                "name",
                "tech",
                "loan",
                "win",
                "pro",
            ]
        )

        # Launch long lived process with signal catcher
        with signal_utils.signal_catcher(save_vtq):

            # Run the zeek reader on the dns.log file looking for risky TLDs
            reader = zeek_log_reader.ZeekLogReader(args.zeek_log)
            for row in reader.readrows():

                # Pull out the TLD
                query = row["query"]
                tld = tldextract.extract(query).suffix

                # Check if the TLD is in the risky group
                if tld in risky_tlds:
                    # Show the risky dns
                    print("Making VT query for {:s}...".format(query))

                    # Make the VT query
                    results = vtq.query_url(query)
                    if results.get("positives", 0) >= 1:  # At least one hit (change this higher if you want)
                        print("\nRisky Domain DNS Query Found")
                        print(
                            "From: {:s} To: {:s} QType: {:s} RCode: {:s}".format(
                                row["id.orig_h"], row["id.resp_h"], row["qtype_name"], row["rcode_name"]
                            )
                        )
                        pprint(results)

        # Save the Virus Total Query
        save_vtq()



"""LogToDataFrame: Converts a Zeek log to a Polars DataFrame"""

# Third Party
from typing import Dict, Optional, List, Tuple

try:
    import polars as pl
except ImportError:
    print("\npip install polars")


# Local
from zat.utils.field_info import get_field_info


class LogToPolars:
    """LogToPolats: Converts a Zeek log to a Polars DataFrame"""

    def __init__(self):
        # Polars data types : https://docs.pola.rs/api/python/stable/reference/datatypes.html
        self.type_map = {
            "str": pl.String,
            "addr": pl.String,
            "table[string]": pl.String,
            "bool": pl.Categorical,
            "double": pl.Float64,
            "time": pl.Float64,
            "interval": pl.Float64,
            "int": pl.Int32,
            "count": pl.UInt64,
            "port": pl.UInt16,
            "enum": pl.Categorical,
        }

    def create_dataframe(self, log_filename: str, usecols: Optional[List[str]] = None) -> "pl.DataFrame":
        """Create a Polars dataframe from a Bro/Zeek log file
        Args:
           log_filename (string): The full path to the Zeek log
           usecol (list): A subset of columns to read in (minimizes memory usage) (default = None)
        """

        # 1. Get field infos.
        field_names, field_types = get_field_info(log_filename=log_filename)
        all_fields = field_names

        # 2. Convert zeek types to polars types.
        #    Replace old types from FieldInfos struct with the converted ones.
        type_map = self._apply_type_map(field_names, field_types)

        # 3. Get only the specified columns
        if usecols:
            field_types = [t for t, field in zip(field_types, field_names) if field in usecols]
            field_names = [field for field in field_names if field in usecols]

        # 3. Get dataframe.
        self._df = self._get_dataframe(
            log_filename=log_filename, all_fields=all_fields, dtypes=type_map, usecols=usecols
        )

        # 4. Convert time type.
        time_cols = [name for name, zt in zip(field_names, field_types) if zt == "time"]
        interval_cols = [name for name, zt in zip(field_names, field_types) if zt == "interval"]

        if time_cols:
            self._df = self._df.with_columns([pl.from_epoch(pl.col(c), time_unit="s") for c in time_cols])

        if interval_cols:
            self._df = self._df.with_columns([(pl.col(c) * 1000).cast(pl.Duration("ms")) for c in interval_cols])

        return self._df

    def _get_dataframe(
        self, log_filename: str, all_fields: List[str], dtypes: Dict, usecols: Optional[List[str]] = None
    ) -> "pl.DataFrame":
        """Internal Method: Create the initial dataframes by using Polars read CSV (primary types correct)"""
        return pl.read_csv(
            log_filename,
            separator="\t",
            has_header=False,
            new_columns=all_fields,
            columns=usecols,
            schema_overrides=dtypes,
            comment_prefix="#",
            null_values=["-", "NA", ""],
        )

    def _apply_type_map(self, column_names: List[str], column_types: List[str]) -> Tuple[List[str], List[str]]:
        polars_types_map = {}
        for name, zeek_type in zip(column_names, column_types):

            # Grab the type
            item_type = self.type_map.get(zeek_type)

            # Sanity Check
            if not item_type:
                item_type = pl.String

            # Set the pandas type
            polars_types_map[name] = item_type

        return polars_types_map


# Simple test of the functionality
def test():
    """Test for LogToDataFrame Class"""
    import os

    import pytest

    try:
        import polars  # noqa: F401
    except ImportError:
        pytest.skip("pip install polars")

    from zat.utils import file_utils

    # Grab a test file
    data_path = file_utils.relative_dir(__file__, "../data")
    log_path = os.path.join(data_path, "conn.log")

    # create_dataframe it to a Pandas DataFrame
    log_to_df = LogToPolars()
    my_df = log_to_df.create_dataframe(log_path)

    # Print out the head
    print(my_df.head())

    # Print out the datatypes
    print(my_df.dtypes)

    # Test a bunch
    tests = [
        "app_stats.log",
        "dns.log",
        "http.log",
        "notice.log",
        "tor_ssl.log",
        "conn.log",
        "dhcp_002.log",
        "files.log",
        "smtp.log",
        "weird.log",
        "ftp.log",
        "ssl.log",
        "x509.log",
    ]
    for log_path in [os.path.join(data_path, log) for log in tests]:
        print("Testing: {:s}...".format(log_path))
        my_df = log_to_df.create_dataframe(log_path)
        print(my_df.head())
        print(my_df.dtypes)

    # Test an empty log (a log with header/close but no data rows)
    # log_path = os.path.join(data_path, 'http_empty.log')
    # my_df = log_to_df.create_dataframe(log_path)

    # Print out the head
    print(my_df.head())

    # Print out the datatypes
    print(my_df.dtypes)

    print("LogToDataFrame Test successful!")



"""Zeek log to Pandas Dataframe + Filter by WhiteList Example"""

import argparse
import os
import sys

import pandas as pd

# Local imports
from zat.log_to_dataframe import LogToDataFrame

if __name__ == "__main__":
    # Example to populate a Pandas dataframe from a zeek log reader

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument("dns_log", help="Specify the zeek DNS log")
    parser.add_argument("whitelist", help="Specify the DNS whiteliist")
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print("Unrecognized args: %s" % commands)
        sys.exit(1)

    # Grab the whitelist
    white_df = pd.read_csv(args.whitelist, names=["rank", "domain"])
    whitelist = white_df["domain"].tolist()

    # File may have a tilde in it
    if args.dns_log:
        args.dns_log = os.path.expanduser(args.dns_log)

        # Create a Pandas dataframe from a Zeek log
        log_to_df = LogToDataFrame()
        zeek_df = log_to_df.create_dataframe(args.dns_log)

        # Print out the head of the dataframe
        print("DF Size before whitelist: {:d} rows".format(len(zeek_df)))

        # Filter the dataframe with the whitelist
        white_df = zeek_df[zeek_df["query"].isin(whitelist)]
        print("Filtering out {!r}".format(white_df["query"].tolist()))
        zeek_df = zeek_df[~zeek_df["query"].isin(whitelist)]

        print("DF Size after whitelist: {:d} rows".format(len(zeek_df)))



"""Risky DNS zat Example"""

import argparse
import json
import pickle
import sys
from pprint import pprint

from kafka import KafkaConsumer
from kafka.errors import NoBrokersAvailable

# Third Party Imports
try:
    import tldextract
except ImportError:
    print("\nThis example needs tldextract. Please do a $pip install tldextract and rerun this example")
    sys.exit(1)

# Local imports
from zat.utils import signal_utils, vt_query


def save_vtq():
    """Exit on Signal"""
    global vtq  # noqa: F824

    print("Saving VirusTotal Query Cache...")
    pickle.dump(vtq, open("vtq.pkl", "wb"), protocol=pickle.HIGHEST_PROTOCOL)
    sys.exit()


if __name__ == "__main__":
    # Risky DNS/VT Query application

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--server", type=str, default="localhost:9092", help="Specify the Kafka Server (default: localhost:9092)"
    )
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print("Unrecognized args: %s" % commands)
        sys.exit(1)

    # First we create a Kafka Consumer
    kserver = args.server
    try:
        consumer = KafkaConsumer(
            "dns", bootstrap_servers=[kserver], value_deserializer=lambda x: json.loads(x.decode("utf-8"))
        )
    except NoBrokersAvailable:
        print("Could not connect to Kafka server: {:s}".format(args.server))
        sys.exit(-1)

    # See if we have a serialized VirusTotal Query Class.
    # If we do not have one we'll create a new one
    try:
        vtq = pickle.load(open("vtq.pkl", "rb"))
        print("Opening VirusTotal Query Cache (cache_size={:d})...".format(vtq.size))
    except IOError:
        vtq = vt_query.VTQuery(max_cache_time=60 * 24 * 7)  # One week cache

    # See our 'Risky Domains' Notebook for the analysis and
    # statistical methods used to compute this risky set of TLDs
    risky_tlds = set(
        [
            "info",
            "tk",
            "xyz",
            "online",
            "club",
            "ru",
            "website",
            "in",
            "ws",
            "top",
            "site",
            "work",
            "biz",
            "name",
            "tech",
            "loan",
            "win",
            "pro",
        ]
    )

    # Launch long lived process with signal catcher
    with signal_utils.signal_catcher(save_vtq):

        # Now lets process our Kafka 'dns' Messages
        for message in consumer:
            dns_message = message.value

            # Pull out the TLD
            query = dns_message.get("query")
            tld = tldextract.extract(query).suffix if query else None

            # Check if the TLD is in the risky group
            if tld in risky_tlds:
                print("\n" + query)
                # Make the query with the full query
                results = vtq.query_url(query)
                if results.get("positives", 0) > 3:  # At least four hits
                    print("Risky Domain DNS Query Found")
                    print(
                        "From: {:s} To: {:s} QType: {:s} RCode: {:s}".format(
                            dns_message["id.orig_h"],
                            dns_message["id.resp_h"],
                            dns_message["qtype_name"],
                            dns_message["rcode_name"],
                        )
                    )
                    pprint(results)



"""Read Kafka Streams into Spark, perform simple filtering/aggregation"""

import argparse
import sys
from time import sleep

try:
    from pyspark.sql import SparkSession
    from pyspark.sql.functions import col, from_json, udf
    from pyspark.sql.types import BooleanType, IntegerType, StringType, StructType
except ImportError:
    print("\npip install pyspark")
    sys.exit(1)

try:
    import tldextract
except ImportError:
    print("\nThis example needs tldextract. Please do a $pip install tldextract and rerun this example")
    sys.exit(1)


def exit_program():
    """Exit on Signal"""
    print("Exiting Program...")
    sys.exit()


def compute_domain(query):
    # Pull out the domain
    if query.endswith(".local"):
        return "local"
    return tldextract.extract(query).registered_domain if query else None


if __name__ == "__main__":
    """Read Kafka Streams into Spark, perform simple filtering/aggregation"""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--server", type=str, default="localhost:9092", help="Specify the Kafka Server (default: localhost:9092)"
    )
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print("Unrecognized args: %s" % commands)
        sys.exit(1)

    # Grab the Kafka server
    kserver = args.server

    # Spin up a local Spark Session (with 4 executors)
    spark = (
        SparkSession.builder.master("local[4]")
        .appName("my_awesome")
        .config("spark.jars.packages", "org.apache.spark:spark-sql-kafka-0-10_2.11:2.4.4")
        .getOrCreate()
    )
    spark.sparkContext.setLogLevel("ERROR")

    # Optimize the conversion to Spark
    spark.conf.set("spark.sql.execution.arrow.enable", "true")

    # SUBSCRIBE: Setup connection to Kafka Stream
    raw_data = (
        spark.readStream.format("kafka")
        .option("kafka.bootstrap.servers", kserver)
        .option("subscribe", "dns")
        .option("startingOffsets", "earliest")
        .load()
    )

    # Define the schema for the DNS message (do this better)
    dns_schema = (
        StructType()
        .add("ts", StringType())
        .add("uid", StringType())
        .add("id.orig_h", StringType())
        .add("id.orig_p", IntegerType())
        .add("id.resp_h", StringType())
        .add("id.resp_p", IntegerType())
        .add("proto", StringType())
        .add("trans_id", IntegerType())
        .add("query", StringType())
        .add("qclass", IntegerType())
        .add("qclass_name", StringType())
        .add("qtype", IntegerType())
        .add("qtype_name", StringType())
        .add("rcode", IntegerType())
        .add("rcode_name", StringType())
        .add("AA", BooleanType())
        .add("TC", BooleanType())
        .add("RD", BooleanType())
        .add("RA", BooleanType())
        .add("Z", IntegerType())
        .add("answers", StringType())
        .add("TTLs", StringType())
        .add("rejected", BooleanType())
    )

    # ETL: Convert raw data into parsed and proper typed data
    parsed_data = raw_data.select(from_json(col("value").cast("string"), dns_schema).alias("data")).select("data.*")

    # FILTER: Only get DNS records that have 'query' field filled out
    filtered_data = parsed_data.filter(parsed_data.query.isNotNull() & (parsed_data.query != ""))

    # FILTER 2: Remove Local/mDNS queries
    filtered_data = filtered_data.filter(~filtered_data.query.like("%.local"))  # Note: using the '~' negation operator

    # COMPUTE: A new column with the 2nd level domain extracted from the query
    udf_compute_domain = udf(compute_domain, StringType())
    computed_data = filtered_data.withColumn("domain", udf_compute_domain("query"))

    # AGGREGATE: In this case a simple groupby operation
    group_data = computed_data.groupBy("`id.orig_h`", "domain", "qtype_name").count()

    # At any point in the pipeline you can see what you're getting out
    group_data.printSchema()

    # Take the end of our pipeline and pull it into memory
    dns_count_memory_table = (
        group_data.writeStream.format("memory").queryName("dns_counts").outputMode("complete").start()
    )

    # Let the pipeline pull some data
    print("Pulling pipline...Please wait...")

    # Create a Pandas Dataframe by querying the in memory table and converting
    # Loop around every 5 seconds to update output
    for _ in range(10):
        sleep(5)
        dns_counts_df = spark.sql("select * from dns_counts").toPandas()
        print("\nDNS Query Total Counts = {:d}".format(dns_counts_df["count"].sum()))
        print(dns_counts_df.sort_values(ascending=False, by="count"))

    # Stop the stream
    dns_count_memory_table.stop()
    sleep(1)



"""NGram utilities that might be useful"""


def compute_ngrams(word_list, S=3, T=3):
    """Compute NGrams in the word_list from [S-T)
    Args:
        word_list (list): A list of words to compute ngram set from
        S (int): The smallest NGram (default=3)
        T (int): The biggest NGram (default=3)
    """
    _ngrams = []
    if isinstance(word_list, str):
        word_list = [word_list]
    for word in word_list:
        for n in range(S, T + 1):
            _ngrams += zip(*(word[i:] for i in range(n)))
    return ["".join(_ngram) for _ngram in _ngrams]


def ngram_count(word, ngrams):
    """Compute the number of matching NGrams in the given word"""
    return len(set(ngrams).intersection(compute_ngrams([word])))


def test():
    """Test the ngram methods"""
    domains = ["google", "facebook", "apple"]
    compute_ngrams(domains, 2, 5)
    ngrams = compute_ngrams(domains)
    print("NGrams: {:s}".format(str(list(ngrams))))
    print(ngram_count("foogle", ngrams))
    print(ngram_count("mybook", ngrams))


if __name__ == "__main__":
    test()



"""Zeek log to Pandas Dataframe Example"""

import argparse
import os
import sys

# Local imports
from zat.log_to_dataframe import LogToDataFrame

if __name__ == "__main__":
    # Example to populate a Pandas dataframe from a zeek log reader

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument("zeek_log", type=str, help="Specify a zeek log to run ZeekLogReader test on")
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print("Unrecognized args: %s" % commands)
        sys.exit(1)

    # File may have a tilde in it
    if args.zeek_log:
        args.zeek_log = os.path.expanduser(args.zeek_log)

        # Create a Pandas dataframe from a Zeek log
        log_to_df = LogToDataFrame()
        zeek_df = log_to_df.create_dataframe(args.zeek_log)

        # Print out the head of the dataframe
        print(zeek_df.head())

        # Print out the types of the columns
        print(zeek_df.dtypes)

        # Print out size and memory usage
        print("DF Shape: {:s}".format(str(zeek_df.shape)))
        print("DF Memory:")
        memory_usage = zeek_df.memory_usage(deep=True)
        total = memory_usage.sum()
        for item in memory_usage.items():
            print("\t {:s}: \t{:.2f} MB".format(item[0], item[1] / 1e6))
        print("DF Total: {:.2f} GB".format(total / (1e9)))



"""LogToDataFrame: Converts a Zeek log to a Dask DataFrame"""

from typing import Dict, List, Optional, Tuple

try:
    import dask.dataframe as dd
except ImportError:
    print("\npip install dask")

from zat.utils.field_info import get_field_info


class LogToDask:
    """LogToDask: Converts a Zeek log to a Dask DataFrame"""

    def __init__(self):
        self.type_map = {
            "bool": "category",  # Can't hold NaN values in 'bool', so we're going to use category
            "count": "UInt64",
            "int": "Int32",
            "double": "float",
            "time": "float",  # Secondary Processing into datetime
            "interval": "float",  # Secondary processing into timedelta
            "port": "UInt16",
            "addr": "string",
        }

    def create_dataframe(
        self,
        log_filename: str,
        ts_index: bool = True,
        aggressive_category: bool = True,
        usecols: Optional[List[str]] = None,
    ) -> "dd.DataFrame":
        """Create a Dask dataframe from a Bro/Zeek log file
        Args:
           log_filename (string): The full path to the Zeek log
           ts_index (bool): Set the index to the 'ts' field (default = True)
           aggressive_category (bool): convert unknown columns to category (default = True)
           usecol (list): A subset of columns to read in (minimizes memory usage) (default = None)
        """

        # Grab the field information
        field_names, field_types = get_field_info(log_filename=log_filename)
        all_fields = field_names  # We need ALL the fields for later

        # If usecols is set then we'll subset the fields and types
        if usecols:
            # Usecols needs to include ts
            if "ts" not in usecols:
                usecols.append("ts")
            field_types = [t for t, field in zip(field_types, field_names) if field in usecols]
            field_names = [field for field in field_names if field in usecols]

        # Get the appropriate types for the dask Dataframe
        dask_types = self._apply_type_map(field_names, field_types, aggressive_category)

        # Now actually read in the initial dataframe
        self._df = self._get_dataframe(
            log_filename=log_filename, all_fields=all_fields, dtypes=dask_types, usecols=usecols
        )

        # Now we convert 'time' and 'interval' fields to datetime and timedelta respectively
        for name, zeek_type in zip(field_names, field_types):
            if zeek_type == "time":
                self._df[name] = dd.to_datetime(self._df[name], unit="s").dt.floor("us")
            if zeek_type == "interval":
                self._df[name] = dd.to_timedelta(self._df[name]).dt.total_seconds()

        # Set the index
        # .empty isn't supported by dask. This condition is a workaround
        if ts_index and len(self._df.columns) > 0:
            try:
                self._df.set_index("ts", inplace=True)
            except KeyError:
                print("Could not find ts/timestamp for index...")
        return self._df

    def _get_dataframe(
        self, log_filename: str, all_fields: List[str], dtypes: Dict, usecols: Optional[List[str]]
    ) -> "dd.DataFrame":
        """Internal Method: Create the initial dataframes by using dask read CSV (primary types correct)"""
        return dd.read_csv(
            log_filename, sep="\t", names=all_fields, usecols=usecols, dtype=dtypes, comment="#", na_values="-"
        )

    def _apply_type_map(
        self, column_names: List[str], column_types: List[str], aggressive_category: bool = True, verbose: bool = False
    ) -> Tuple[List[str], List[str]]:
        """Given a set of names and types, construct a Dictionary to be used
        as the dask read_csv dtypes argument"""

        # Aggressive Category means that types not in the current type_map are
        # mapped to a 'category' if aggressive_category is False then they
        # are mapped to an 'object' type
        unknown_type = "category" if aggressive_category else "object"

        dask_types = {}
        for name, zeek_type in zip(column_names, column_types):

            # Grab the type
            item_type = self.type_map.get(zeek_type)

            # Sanity Check
            if not item_type:
                # UID/FUID/GUID always gets mapped to object
                if "uid" in name:
                    item_type = "object"
                else:
                    if verbose:
                        print("Could not find type for {:s} using {:s}...".format(zeek_type, unknown_type))
                    item_type = unknown_type

            # Set the dask type
            dask_types[name] = item_type

        # Return the dictionary of name: type
        return dask_types


# Simple test of the functionality
def test():
    """Test for LogToDataFrame Class"""
    import os

    import pytest

    try:
        import dask.dataframe  # noqa: F401
    except ImportError:
        pytest.skip("pip install dask")

    from zat.utils import file_utils

    # Grab a test file
    data_path = file_utils.relative_dir(__file__, "../data")
    log_path = os.path.join(data_path, "conn.log")

    # create_dataframe it to a Pandas DataFrame
    log_to_df = LogToDask()
    my_df = log_to_df.create_dataframe(log_path)

    # Print out the head
    print(my_df.head())

    # Print out the datatypes
    print(my_df.dtypes)

    # Test a bunch
    tests = [
        "app_stats.log",
        "dns.log",
        "http.log",
        "notice.log",
        "tor_ssl.log",
        "conn.log",
        "dhcp_002.log",
        "files.log",
        "smtp.log",
        "weird.log",
        "ftp.log",
        "ssl.log",
        "x509.log",
    ]
    for log_path in [os.path.join(data_path, log) for log in tests]:
        print("Testing: {:s}...".format(log_path))
        my_df = log_to_df.create_dataframe(log_path)
        print(my_df.head())
        print(my_df.dtypes)

    # Test out usecols arg
    conn_path = os.path.join(data_path, "conn.log")
    my_df = log_to_df.create_dataframe(
        conn_path, usecols=["id.orig_h", "id.orig_p", "id.resp_h", "id.resp_p", "proto", "orig_bytes", "resp_bytes"]
    )

    # Test an empty log (a log with header/close but no data rows)
    log_path = os.path.join(data_path, "http_empty.log")
    my_df = log_to_df.create_dataframe(log_path)

    # Print out the head
    print(my_df.head())

    # Print out the datatypes
    print(my_df.dtypes)

    print("LogToDataFrame Test successful!")



"""LogToSparkDF: Converts a Zeek log to a Spark DataFrame"""

# Third Party
try:
    from pyspark.sql.functions import col, when
    from pyspark.sql.types import DoubleType, FloatType, IntegerType, LongType, StringType, StructType
except ImportError:
    print("\npip install pyspark")


# Local
from zat.utils.field_info import get_field_info


class LogToSparkDF:
    """LogToSparkDF: Converts a Zeek log to a Spark DataFrame"""

    def __init__(self, spark):
        """Initialize the LogToSparkDF class"""

        # Grab the spark context
        self.spark = spark

        # First Level Type Mapping
        #    This map defines the types used when first reading in the Zeek log into a 'chunk' dataframes.
        #    Types (like time and interval) will be defined as one type at first but then
        #    will undergo further processing to produce correct types with correct values.
        # See: https://spark.apache.org/docs/latest/sql-reference.html
        #      for more info on supported types.
        self.type_map = {
            "bool": StringType(),  # Secondary Processing into BooleanType()
            "count": LongType(),
            "int": IntegerType(),
            "double": FloatType(),
            "time": DoubleType(),  # Secondary Processing into TimestampType()
            "interval": FloatType(),
            "port": IntegerType(),
            "enum": StringType(),
            "addr": StringType(),
            "string": StringType(),
        }

    def create_dataframe(self, log_filename, fillna=True):
        """Create a Spark dataframe from a Bro/Zeek log file
        Args:
           log_fllename (string): The full path to the Zeek log
           fillna (bool): Fill in NA/NaN values (default=True)
        """

        # Create a Zeek log reader just to read in the header for names and types
        field_names, field_types = get_field_info(log_filename=log_filename)

        spark_schema = self.build_spark_schema(column_names=field_names, column_types=field_types)

        # Now actually read the Zeek Log using Spark read CSV
        _df = self.spark.read.csv(log_filename, schema=spark_schema, sep="\t", comment="#", nullValue="-")

        """ Secondary processing (cleanup)
            - Fix column names with '.' in them
            - Fill in Nulls (optional)
            - timestamp convert
            - boolean convert
        """

        # Fix column names
        """ Note: Yes column names with '.' in them can be escaped with backticks when selecting them BUT
                  many pipeline operations will FAIL internally if the column names have a '.' in them.
        """
        fixed_columns = list(map(lambda x: x.replace(".", "_"), _df.columns))
        _df = _df.toDF(*fixed_columns)

        # Fill in NULL values
        if fillna:
            _df = _df.na.fill(0)  # For numeric columns
            _df = _df.na.fill("-")  # For string columns

        # Convert timestamp and boolean columns
        for name, f_type in zip(field_names, field_types):
            # Some field names may have '.' in them, so we create a reference name to those fields
            ref_name = name.replace(".", "_")
            if f_type == "time":
                _df = _df.withColumn(name, _df[ref_name].cast("timestamp"))
            if f_type == "bool":
                _df = _df.withColumn(
                    name,
                    when(col(ref_name) == "T", "true")
                    .when(col(ref_name) == "F", "false")
                    .otherwise("null")
                    .cast("boolean"),
                )

        # Return the spark dataframe
        return _df

    def build_spark_schema(self, column_names, column_types, verbose=False):
        """Given a set of names and types, construct a dictionary to be used
        as the Spark read_csv dtypes argument"""

        # If we don't know the type put it into a string
        unknown_type = StringType()

        schema = StructType()
        for name, zeek_type in zip(column_names, column_types):

            # Grab the type
            spark_type = self.type_map.get(zeek_type)

            # Sanity Check
            if not spark_type:
                if verbose:
                    print("Could not find type for {:s} using StringType...".format(zeek_type))
                spark_type = unknown_type

            # Add the Spark type for this column
            schema.add(name, spark_type)

        # Return the Spark schema
        return schema


# Simple test of the functionality
def test():
    """Test for LogToSparkDF Class"""
    import os

    import pytest

    from zat.utils import file_utils

    pytest.skip("FIXME: Spark 4.x casting changes break this test — skip for now")

    try:
        from pyspark.sql import SparkSession
    except ImportError:
        pytest.skip("pip install pyspark")

    # Spin up a local Spark Session (with 4 executors)
    spark = SparkSession.builder.master("local[4]").appName("my_awesome").getOrCreate()

    # Grab a test file
    data_path = file_utils.relative_dir(__file__, "../data")
    log_path = os.path.join(data_path, "ftp.log")

    # Convert it to a Spark DataFrame
    log_to_spark = LogToSparkDF(spark)
    spark_df = log_to_spark.create_dataframe(log_path)

    # Print out the head
    print(spark_df.show())

    # Print out the datatypes
    print(spark_df.printSchema())

    num_rows = spark_df.count()
    print("Number of Spark DataFrame rows: {:d}".format(num_rows))
    columns = spark_df.columns
    print("Columns: {:s}".format(",".join(columns)))

    # Test a bunch
    tests = [
        "app_stats.log",
        "dns.log",
        "http.log",
        "notice.log",
        "tor_ssl.log",
        "conn.log",
        "dhcp.log",
        "dhcp_002.log",
        "files.log",
        "smtp.log",
        "weird.log",
        "ftp.log",
        "ssl.log",
        "x509.log",
    ]
    for log_path in [os.path.join(data_path, log) for log in tests]:
        print("Testing: {:s}...".format(log_path))
        spark_df = log_to_spark.create_dataframe(log_path)
        print(spark_df.show())
        print(spark_df.printSchema())

    # Test an empty log (a log with header/close but no data rows)
    log_path = os.path.join(data_path, "http_empty.log")
    spark_df = log_to_spark.create_dataframe(log_path)
    print(spark_df.show())
    print(spark_df.printSchema())

    print("LogToSparkDF Test successful!")


if __name__ == "__main__":
    # Run the test for easy testing/debugging
    test()



"""KafkaRouter: The class takes in N Kafka input topics and produces M Kafka output topics"""

import json
import sys
from collections import defaultdict
from datetime import datetime

from kafka import KafkaConsumer, KafkaProducer
from kafka.errors import NoBrokersAvailable

# Local imports
from zat.utils import signal_utils


class KafkaRouter(object):
    """KafkaRouter: The class takes in N Kafka input topics and produces M Kafka output topics

    Args:
        input_topics (list of strings): List of the initial input topics
        input_servers (list of strings): Kafka Bootstrap servers (default=['localhost:9092'])
        output_servers (list of strings): Kafka Bootstrap servers (default=['localhost:9092'])
    """

    def __init__(self, offset="latest", input_servers=["localhost:9092"], output_servers=["localhost:9092"]):
        """Initialize the KafkaRouter class"""

        # Setup the input and output
        print("Initializing KafkaRouter: {!r}".format(input_servers))
        try:
            self.input_pipe = KafkaConsumer(
                bootstrap_servers=input_servers,
                auto_offset_reset=offset,
                value_deserializer=lambda x: json.loads(x.decode("utf8")),
            )
        except NoBrokersAvailable:
            print("Could not connect to Kafka bootstrap servers: {:s}".format(input_servers))
            sys.exit(-1)
        try:
            self.output_pipe = KafkaProducer(
                bootstrap_servers=output_servers, value_serializer=lambda x: json.dumps(x).encode("utf8")
            )
        except NoBrokersAvailable:
            print("Could not connect to Kafka bootstrap servers: {:s}".format(output_servers))
            sys.exit(-1)

        # Gates
        self.route_info = {}
        self.routes = defaultdict(list)

        # Topics we're listening to
        self.topics = set()

    def add_route(self, topic, callback):
        """Add a logic route that pulls in a message from the input_pipe and sends messages
        to the output_pipe with the specified topic"""
        self.route_info[callback.__name__] = callback
        self.routes[topic].append(callback)

        # Add this topic to our input pipe
        if topic not in self.topics:
            self.topics.add(topic)
            self.input_pipe.subscribe(list(self.topics))
            print("Adding Topic {:s}".format(topic))
            print("Topics: {!r}".format(self.input_pipe.subscription()))

    def run(self):
        """Run the KafkaRouter with all of the registered logic routes"""
        with signal_utils.signal_catcher(self.exit_program):

            # Now lets process our Kafka Messages
            for message in self.input_pipe:
                topic = message.topic
                message = message.value
                for route in self.routes[topic]:
                    topic = route(message)
                    if topic:
                        self.output_pipe.send(topic, message)
                        # self.output_pipe.poll()  # What do we want to do here?

    def list_routes(self):
        print("{!r}".format(self.route_info))

    def exit_program(self):
        """Exit on Signal"""
        print("Exiting Program...")
        sys.exit()


# Simple test of the functionality
def disabled_test():
    """Test for KafkaRouter Class"""
    from zat.utils import geo_lookup

    my_geo = geo_lookup.GeoLookup()

    # Make some simple logic routes
    def north_south(message):
        if (not message["local_orig"]) and message["local_resp"]:
            return "incoming"
        elif message["local_orig"] and not message["local_resp"]:
            return "outgoing"
        return None

    def incoming_info(message):
        geo_info = my_geo.query_ip(message["id.orig_h"])
        timestamp = datetime.fromtimestamp(message["ts"])
        print("\nINCOMING")
        print(
            timestamp,
            geo_info["country_code"],
            geo_info["region_name"],
            message["id.orig_h"],
            "-->",
            message["id.resp_h"],
            message["proto"],
            message.get("service", "unknown"),
        )
        return None

    def outgoing_info(message):
        ip = message["id.resp_h"]
        # Skip broad-cast, multi-cast
        if ip[:3] in ["255", "239", "224"]:
            return None

        # Get Geographical Information and Check for outside US traffic
        geo_info = my_geo.query_ip(message["id.resp_h"])
        if not geo_info or geo_info["country_code"] == "US":
            return None

        # Add geo info and route to 'non_us'
        message["country_code"] = geo_info["country_code"]
        message["region_name"] = geo_info["region_name"]
        return "non_us"

    def print_info(message):
        timestamp = datetime.fromtimestamp(message["ts"])
        print(
            timestamp,
            message["country_code"],
            message["region_name"],
            message["id.orig_h"],
            "-->",
            message["id.resp_h"],
            message["proto"],
            message.get("service", "unknown"),
        )
        return None

    # Create the class and test it
    router = KafkaRouter(offset="earliest")
    router.add_route("conn", north_south)
    router.add_route("outgoing", outgoing_info)
    router.add_route("incoming", incoming_info)
    router.add_route("non_us", print_info)
    router.run()


if __name__ == "__main__":
    disabled_test()



"""Network utilities that might be useful"""

import binascii
import ipaddress
import socket


def mac_to_str(address):
    """Convert a MAC address to a readable/printable string

    Args:
        address (str): a MAC address in hex form (e.g. '\x01\x02\x03\x04\x05\x06')
    Returns:
        str: Printable/readable MAC address
    """
    return ":".join("%02x" % b for b in address)


def str_to_mac(mac_string):
    """Convert a readable string to a MAC address

    Args:
        mac_string (str): a readable string (e.g. '01:02:03:04:05:06')
    Returns:
        str: a MAC address in hex form
    """
    sp = mac_string.split(":")
    mac_string = "".join(sp)
    return binascii.unhexlify(mac_string)


def inet_to_str(inet):
    """Convert inet object to a string

    Args:
        inet (inet struct): inet network address
    Returns:
        str: Printable/readable IP address
    """
    # First try ipv4 and then ipv6
    try:
        return socket.inet_ntop(socket.AF_INET, inet)
    except ValueError:
        return socket.inet_ntop(socket.AF_INET6, inet)


def str_to_inet(address):
    """Convert an a string IP address to a inet struct

    Args:
        address (str): String representation of address
    Returns:
        inet: Inet network address
    """
    # First try ipv4 and then ipv6
    try:
        return socket.inet_pton(socket.AF_INET, address)
    except socket.error:
        return socket.inet_pton(socket.AF_INET6, address)


def is_internal(ip_address):
    """Determine if the address is an internal ip address"""
    return ipaddress.ip_address(ip_address).is_private


def is_special(ip_address):
    """Determine if the address is SPECIAL
    Note: This is super bad, improve it
    """
    special = {"224.0.0.251": "multicast_dns", "ff02::fb": "multicast_dns"}
    return special[ip_address] if ip_address in special else False


def traffic_direction(conn_row):
    """Determine the direction of the connection traffic (takes a conn.log row)"""

    # First try to use the local orig/resp fields
    if conn_row.get("local_orig") and conn_row.get("local_resp"):
        local_orig = conn_row["local_orig"]
        local_resp = conn_row["local_resp"]
    else:
        # Well we don't have local orig/resp fields so use RFC1918 logic
        local_orig = ipaddress.ip_address(conn_row["id.orig_h"]).is_private
        local_resp = ipaddress.ip_address(conn_row["id.resp_h"]).is_private

    # Determine north/south or internal traffic
    if (not local_orig) and local_resp:
        return "incoming"
    if local_orig and not local_resp:
        return "outgoing"

    # Neither host is in the allocated private ranges
    if (
        ipaddress.ip_address(conn_row["id.orig_h"]).is_multicast
        or ipaddress.ip_address(conn_row["id.resp_h"]).is_multicast
    ):
        return "multicast"

    # Both hosts are internal
    return "internal"


def test_utils():
    """Test the utility methods"""

    print(mac_to_str(b"\x01\x02\x03\x04\x05\x06"))
    assert mac_to_str(b"\x01\x02\x03\x04\x05\x06") == "01:02:03:04:05:06"
    assert str_to_mac("01:02:03:04:05:06") == b"\x01\x02\x03\x04\x05\x06"
    my_mac = b"\x01\x02\x03\x04\x05\x06"
    my_str = mac_to_str(my_mac)
    assert str_to_mac(my_str) == my_mac
    print(inet_to_str(b"\x91\xfe\xa0\xed"))
    assert inet_to_str(b"\x91\xfe\xa0\xed") == "145.254.160.237"
    assert str_to_inet("145.254.160.237") == b"\x91\xfe\xa0\xed"

    # IPV6 for Google DNS server
    google_dns = "2001:4860:4860::8888"
    google_inet = b" \x01H`H`\x00\x00\x00\x00\x00\x00\x00\x00\x88\x88"
    assert str_to_inet(google_dns) == google_inet

    # Various methods
    assert is_internal("10.0.0.1")
    assert is_internal("222.2.2.2") is False
    assert is_special("224.0.0.251")
    assert is_special("224.0.0.252") is False
    print("Success!")


if __name__ == "__main__":
    test_utils()



from zat.zeek_log_reader import ZeekLogReader
from typing import Tuple, List


def get_field_info(log_filename: str) -> Tuple[List[str], List[str]]:
    _zeek_reader = ZeekLogReader(log_filename)
    _, field_names, field_types, _ = _zeek_reader._parse_zeek_header(log_filename)
    return field_names, field_types



"""FileTailer Python Class"""

import argparse
import os
import sys
from pprint import pprint

# Local imports
from zat import zeek_log_reader

if __name__ == "__main__":
    # Example to run the zeek log reader on a given file

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument("zeek_log", type=str, help="Specify a zeek log to run ZeekLogReader test on")
    parser.add_argument("-t", "--tail", action="store_true", help="Turn on log tailing")
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print("Unrecognized args: %s" % commands)
        sys.exit(1)

    # File may have a tilde in it
    if args.zeek_log:
        args.zeek_log = os.path.expanduser(args.zeek_log)

        # Run the zeek reader on a given log file
        reader = zeek_log_reader.ZeekLogReader(args.zeek_log, tail=args.tail, strict=True)
        for row in reader.readrows():
            pprint(row)



"""Plotting utilities"""


def plot_defaults():
    try:
        import matplotlib.pyplot as plt

        plt.rc("font", size=14)
        plt.rc("xtick", labelsize=12)
        plt.rc("ytick", labelsize=12)
        plt.rc("axes", labelsize=14)
        plt.rc("axes", axisbelow=True)
        plt.rc("grid", color="grey")
        plt.rc("grid", alpha=0.5)
        plt.rc("grid", linestyle="-")
        plt.rcParams["figure.figsize"] = 16.0, 6.0
        # plt.rcParams.update({'figure.autolayout': True})
        plt.rc("patch", force_edgecolor=True)
    except ImportError:
        print("Could not import matplotlib... this is fine...")


def test():
    """Test the Plot Utilities"""
    plot_defaults()


if __name__ == "__main__":
    test()



"""Read Kafka Streams and Print out the messages"""

import argparse
import json
import sys
from pprint import pprint

from kafka import KafkaConsumer
from kafka.errors import NoBrokersAvailable

# Local imports
from zat.utils import signal_utils


def exit_program():
    """Exit on Signal"""
    print("Exiting Program...")
    sys.exit()


if __name__ == "__main__":
    """Read Kafka Streams and Print out the messages"""

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--server", type=str, default="localhost:9092", help="Specify the Kafka Server (default: localhost:9092)"
    )
    parser.add_argument(
        "--topics",
        type=lambda s: s.split(","),
        default="all",
        help="Specify the Kafka Topics (e.g. dns   or   dns, http, blah   (defaults to all)",
    )
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print("Unrecognized args: %s" % commands)
        sys.exit(1)

    # Create a Kafka Consumer and subscribe to the topics
    all_topics = ["capture_loss", "dns", "http", "ssl", "weird", "conn", "files", "x509"]
    kserver = args.server
    topics = args.topics if args.topics != ["all"] else all_topics
    print("Subscribing to: {!r}".format(topics))
    try:
        consumer = KafkaConsumer(
            *topics, bootstrap_servers=[kserver], value_deserializer=lambda x: json.loads(x.decode("utf-8"))
        )
    except NoBrokersAvailable:
        print("Could not connect to Kafka server: {:s}".format(args.server))
        sys.exit(-1)

    # Launch long lived process with signal catcher
    with signal_utils.signal_catcher(exit_program):

        # Now lets process our Kafka Messages
        for message in consumer:
            topic = message.topic
            message = message.value
            print("\n{:s}".format(topic.upper()))
            pprint(message)
